/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: cmi@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * --------------------------------------------------------------------------
 * $Id: DefaultPolicyStrategyManager.java 2128 2008-10-02 19:24:24Z loris $
 * --------------------------------------------------------------------------
 */

package org.ow2.cmi.controller.server.manager;

import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import org.ow2.cmi.controller.server.AbsServerClusterViewManager;
import org.ow2.cmi.controller.server.IPolicyStrategyManager;
import org.ow2.cmi.controller.server.ServerClusterViewManager;
import org.ow2.cmi.controller.server.ServerClusterViewManagerException;
import org.ow2.cmi.lb.loader.ArchiveId;
import org.ow2.cmi.lb.policy.FirstAvailable;
import org.ow2.cmi.lb.policy.HASingleton;
import org.ow2.cmi.lb.policy.IPolicy;
import org.ow2.cmi.lb.policy.Random;
import org.ow2.cmi.lb.policy.RoundRobin;
import org.ow2.cmi.lb.strategy.IStrategy;
import org.ow2.cmi.lb.strategy.LoadFactorComparator;
import org.ow2.cmi.lb.strategy.LoadFactorSort;
import org.ow2.cmi.lb.strategy.LocalPreference;
import org.ow2.cmi.lb.strategy.NoStrategy;
import org.ow2.cmi.reference.CMIReference;
import org.ow2.cmi.reference.ObjectNotFoundException;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Default manager for policies and strategies.
 *
 * @author The new CMI team
 *
 */
public class DefaultPolicyStrategyManager implements IPolicyStrategyManager {

    /**
     * Logger.
     */
    private static final Log LOGGER = LogFactory
            .getLog(AbsServerClusterViewManager.class);

    /**
     * Embedded Policy classes.
     */
    private final Class<?>[] embeddedPolicyClasses =
    {FirstAvailable.class, HASingleton.class, Random.class, RoundRobin.class};

    /**
     * Embedded Strategy classes.
     */
    private final Class<?>[] embeddedStrategyClasses =
    {NoStrategy.class, LoadFactorSort.class, LoadFactorComparator.class, LocalPreference.class};

    /**
     * Store each class of Policy.
     */
    @SuppressWarnings("unchecked")
    private final ConcurrentHashMap<String, Class<? extends IPolicy>> policyMap =
        new ConcurrentHashMap<String, Class<? extends IPolicy>>();

    /**
     * Store each class of Strategy.
     */
    @SuppressWarnings("unchecked")
    private final ConcurrentHashMap<String, Class<? extends IStrategy>> strategyMap =
        new ConcurrentHashMap<String, Class<? extends IStrategy>>();

    /**
     * Store default strategy name for each object.
     */
    private final ConcurrentHashMap<String, String> defaultStrategyMap = new ConcurrentHashMap<String, String>();

    /**
     * Store default policy name for each object.
     */
    private final ConcurrentHashMap<String, String> defaultPolicyMap = new ConcurrentHashMap<String, String>();

    /**
     * Current Server Cluster View Manager.
     */
    private ServerClusterViewManager serverClusterViewManager = null;

    /**
     * Constructor without fields.
     */
    public DefaultPolicyStrategyManager() {
    }

    /**
     * Constructor using fields.
     *
     * @param srvClusterViewManager
     *            the server-side manager to communicate with.
     */
    public DefaultPolicyStrategyManager(
            final ServerClusterViewManager srvClusterViewManager) {
        this.serverClusterViewManager = srvClusterViewManager;
    }

    /**
     * Returns the class that defines the policy for an object with the given
     * name.
     *
     * @param objectName
     *            a name of object
     * @return the class that defines the policy for an object with the given
     *         name
     * @throws ObjectNotFoundException
     *             if none object has the given name
     * @throws ServerClusterViewManagerException
     *             if the class cannot be loaded
     */
    @SuppressWarnings("unchecked")
    public final Class<? extends IPolicy<?>> getPolicyClass(
            final String objectName) throws ObjectNotFoundException,
            ServerClusterViewManagerException {

        String policyClassName = serverClusterViewManager
                .getPolicyClassName(objectName);
        if (policyMap.containsKey(policyClassName)) {
            return (Class<? extends IPolicy<?>>) policyMap.get(policyClassName);
        }
        try {
            Class<? extends IPolicy<?>> policyClass = (Class<? extends IPolicy<?>>) Class
                    .forName(policyClassName);
            policyMap.putIfAbsent(policyClassName, policyClass);
            return policyClass;
        } catch (ClassNotFoundException e) {
            LOGGER.error("Cannot load the class for policy " + policyClassName,
                    e);
            throw new ServerClusterViewManagerException(
                    "Cannot load the class for policy " + policyClassName, e);
        }
    }

    /**
     * Returns the class that defines the strategy for an object with the given
     * name.
     *
     * @param objectName
     *            a name of object
     * @return the class that defines the strategy for an object with the given
     *         name
     * @throws ObjectNotFoundException
     *             if none object has the given name
     * @throws ServerClusterViewManagerException
     *             if the class cannot be loaded
     */
    @SuppressWarnings("unchecked")
    public final Class<? extends IStrategy<?>> getStrategyClass(
            final String objectName) throws ObjectNotFoundException,
            ServerClusterViewManagerException {
        String strategyClassName = serverClusterViewManager
                .getStrategyClassName(objectName);
        if (strategyMap.containsKey(strategyClassName)) {
            return (Class<? extends IStrategy<?>>) strategyMap
                    .get(strategyClassName);
        }
        try {
            Class<? extends IStrategy<?>> strategyClass = (Class<? extends IStrategy<?>>) Class
                    .forName(strategyClassName);
            strategyMap.putIfAbsent(strategyClassName, strategyClass);
            return strategyClass;
        } catch (ClassNotFoundException e) {
            LOGGER.error("Cannot load the class for strategy "
                    + strategyClassName, e);
            throw new ServerClusterViewManagerException(
                    "Cannot load the class for strategy " + strategyClassName,
                    e);
        }
    }

    /**
     * Receives archive containing load balancing classes. The server cluster
     * view manager delegate the call to the policy/Strategy manager. Since
     * complex treatment might be performed, to assume the reception of the
     * archive.
     *
     * @param bytesOfFile
     *            Bytes content of the file
     * @param location
     *            File location
     * @param strategies
     *            policies exported by the archive
     * @param policies
     *            strategies exported by the archive
     * @return The archive Id used by <code>PolicyStrategyManager</code> to
     *         register the new archive.
     */
    @SuppressWarnings("unchecked")
    public Object addLoadBalancingArchive(final byte[] bytesOfFile,
            final String location, final Class<?>[] policies,
            final Class<?>[] strategies) {
        LOGGER
                .debug(
                        "Sending load balancing archive located at '{0}'. Step 4:"
                                + " policy/strategy manager is sending archive data to be distrubuted in the JGroups cluster",
                        location);
        ArchiveId id = PolicyStrategyManagerUtil.buildArchiveId(
                serverClusterViewManager.getUUID().toString(), location);
        LOGGER.debug("Archive Id is '{0}'", id.toString());
        for (Class<?> policy : policies) {
            policyMap.put(policy.getName(),
                    (Class<? extends IPolicy<?>>) policy);
        }

        for (Class<?> strategy : strategies) {
            strategyMap.put(strategy.getName(),
                    (Class<? extends IStrategy<?>>) strategy);
        }
        /**
         * Another implementation of the policy/strategy manager should make it
         * another way.
         */
        if (!((AbsServerClusterViewManager) serverClusterViewManager)
                .isAlreadyDistributed(id.toString())) {
            ((AbsServerClusterViewManager) serverClusterViewManager)
                    .broadCastArchive(id, bytesOfFile);
        }
        return id.toString();
    }

    /**
     * @return the serverClusterViewManager
     */
    public ServerClusterViewManager getServerClusterViewManager() {
        return serverClusterViewManager;
    }

    /**
     * @param serverClusterViewManager
     *            the serverClusterViewManager to set
     */
    public void setServerClusterViewManager(
            final ServerClusterViewManager serverClusterViewManager) {
        this.serverClusterViewManager = serverClusterViewManager;
    }

    /**
     * Add a policy class.
     *
     * @param policyClass
     *            the policy class to add.
     * @param overwrite
     *            if true then existing class with the given name be
     *            overwritten.
     * @throws ServerClusterViewManagerException
     *             if <code>policyClass</code> cannot be added.
     */
    public void addPolicyClass(final Class<? extends IPolicy<?>> policyClass,
            final boolean overwrite) throws ServerClusterViewManagerException {
        if (overwrite) {
            if (policyMap.contains(policyClass.getName())) {
                policyMap.remove(policyClass.getName());
                policyMap.put(policyClass.getName(), policyClass);
                return;
            }
        }
        policyMap.putIfAbsent(policyClass.getName(), policyClass);
    }

    /**
     * Add a strategy class.
     *
     * @param strategyClass
     *            the strategy class to add.
     * @param overwrite
     *            if true then existing class with the given name be
     *            overwritten.
     * @throws ServerClusterViewManagerException
     *             if <code>strategyClass</code> cannot be added.
     */
    public void addStrategyClass(
            final Class<? extends IStrategy<?>> strategyClass,
            final boolean overwrite) throws ServerClusterViewManagerException {
        if (overwrite) {
            if (strategyMap.contains(strategyClass.getName())) {
                strategyMap.remove(strategyClass.getName());
                strategyMap.put(strategyClass.getName(), strategyClass);
                return;
            }
        }
        strategyMap.putIfAbsent(strategyClass.getName(), strategyClass);
    }

    /**
     * Remove Load Balancing Archive, previously added.
     *
     * @param archiveId
     *            The Id of the archive.
     * @param policies
     *            names of the policies to remove
     * @param strategies
     *            names of the policies to remove
     * @param isArchiveOwner
     *            if true then the call is performed by the archive owner. The
     *            archive owner is the first server that deployed the archive.
     * @throws Exception
     *
     */
    public void removeLoadBalancingArchive(final Object archiveId,
            final String[] policies, final String[] strategies,
            final boolean isArchiveOwner) throws Exception {
        Set<String> objects = serverClusterViewManager.getObjectNames();
        for (String policy : policies) {
            for (Iterator<String> iterator = objects.iterator(); iterator
                    .hasNext();) {
                String objectName = iterator.next();
                String objectPolicyName = serverClusterViewManager
                        .getPolicyClassName(objectName);
                if (objectPolicyName.equals(policy)) {
                    /**
                     * Set default policy
                     */
                    if (isArchiveOwner) {
                        serverClusterViewManager.setPolicyClassName(objectName,
                                getDefaultPolicyClassName(objectName));
                    }
                }
            }
            policyMap.remove(policy);
        }

        objects = serverClusterViewManager.getObjectNames();
        for (String strategy : strategies) {
            for (Iterator<String> iterator = objects.iterator(); iterator
                    .hasNext();) {
                String objectName = iterator.next();
                String objectStrategyName = serverClusterViewManager
                        .getStrategyClassName(objectName).getClass().getName();
                if (objectStrategyName.equals(strategy)) {
                    /**
                     * Set default strategy
                     */
                    if (isArchiveOwner) {
                        serverClusterViewManager.setStrategyClassName(
                                objectName,
                                getDefaultStrategyClassName(objectName));
                    }
                }
            }
            strategyMap.remove(strategy);
        }
        /**
         * The archive will be shared no more.
         */
        if (isArchiveOwner) {
            ((AbsServerClusterViewManager) serverClusterViewManager)
                    .removeDistributedArchive(archiveId);
        }

    }

    /**
     * Get default policy class name of the object.
     *
     * @param objectName
     *            object to get default policy class name.
     * @return
     */
    private String getDefaultPolicyClassName(final String objectName)
            throws PolicyStrategyManagerException {
        if (defaultPolicyMap.containsKey(objectName)) {
            return defaultPolicyMap.get(objectName);
        }
        throw new PolicyStrategyManagerException("No default policy found for "
                + objectName);
    }

    /**
     * Get default strategy class name of the object.
     *
     * @param objectName
     *            object to get default strategy class name.
     * @return Default Strategy ClassName associated to <code>objectName</code>
     * @throws PolicyStrategyManagerException
     *             if no default strategy was found for <code>objectName</code>
     */
    private String getDefaultStrategyClassName(final String objectName)
            throws PolicyStrategyManagerException {
        if (defaultStrategyMap.contains(objectName)) {
            return defaultStrategyMap.get(objectName);
        }
        throw new PolicyStrategyManagerException(
                "No default strategy found for " + objectName);
    }

    /**
     * Add default policy for the <code>objectName</code> parameter.
     *
     * @param objectName
     *            name of the object.
     * @param policyClassName
     *            the policy name
     */
    public void addDefaultPolicyClassName(final String objectName,
            final String policyClassName) {
        LOGGER.debug("Object: {0}, default policy:{1} ", objectName,
                policyClassName);
        defaultPolicyMap.putIfAbsent(objectName, policyClassName);
    }

    /**
     * Add default strategy for the <code>objectName</code> parameter.
     *
     * @param objectName
     *            name of the object.
     * @param strategyClassName
     *            the policy name
     */
    public void addDefaultStrategyClassName(final String objectName,
            final String strategyClassName) {
        LOGGER.debug("Object: {0}, default strategy:{1} ", objectName,
                strategyClassName);
        defaultStrategyMap.putIfAbsent(objectName, strategyClassName);
    }

    /**
     * Get all the policies.
     *
     * @return the policy class names.
     */
    public Collection<? extends String> getPolicyClassNames() {
        return policyMap.keySet();
    }

    /**
     * Get all the strategies.
     *
     * @return the strategy class names.
     */
    public Collection<? extends String> getStrategyClassNames() {
        return strategyMap.keySet();
    }

    /**
     * Load classes embedded in the CMI distribution.
     */
    @SuppressWarnings("unchecked")
    public void loadEmbeddedLBClasses() {
        for (Class<?> embeddedPolicyClass : embeddedPolicyClasses) {
            policyMap.put(embeddedPolicyClass.getName(), (Class<? extends IPolicy<CMIReference>>) embeddedPolicyClass);
        }
        for (Class<?> embeddedStrategyClass : embeddedStrategyClasses) {
            strategyMap.put(embeddedStrategyClass.getName(), (Class<? extends IStrategy<CMIReference>>) embeddedStrategyClass);
        }
    }

    /**
     * Return true if the given class name is embedded policy class.
     * @param className a name of class
     * @return true if the given class name is embedded policy class
     */
    public final boolean isEmbeddedPolicy(final String className) {
        for (Class<?> embeddedPolicyClass : embeddedPolicyClasses) {
            if (embeddedPolicyClass.getName().equals(className)) {
                return true;
            }
        }
        return false;
    }


    /**
     * Return true if the given class name is embedded strategy class.
     * @param className a name of class
     * @return true if the given class name is embedded strategy class
     */
    public final boolean isEmbeddedStrategy(final String className) {
        for (Class<?> embeddedStrategyClass : embeddedStrategyClasses) {
            if (embeddedStrategyClass.getName().equals(className)) {
                return true;
            }
        }
        return false;
    }

}
